from pd_ai_agent_core.parallels_desktop.models.os import Os
from pd_ai_agent_core.parallels_desktop.models.update_package import AppPackage


def parse_linux_info(info_text) -> Os:
    """
    Parse Ubuntu system information text into a dictionary.

    Args:
        info_text (str): The raw text output from lsb_release or similar command

    Returns:
        dict: A dictionary containing the parsed Ubuntu information
    """
    result = {}

    # Split the text into lines and process each line
    for line in info_text.strip().split("\n"):
        # Skip empty lines
        if not line.strip():
            continue

        # Skip the "No LSB modules are available" line
        if "No LSB modules are available" in line:
            continue

        # Split by colon and handle the key-value pairs
        parts = line.split(":", 1)
        if len(parts) == 2:
            key = parts[0].strip()
            value = parts[1].strip()
            result[key] = value
    return Os(
        name=result["Distributor ID"],
        version=result["Release"],
        codename=result["Codename"],
        release=result["Description"],
    )


def parse_windows_info(info_text) -> Os:
    """
    Parse Windows system information text into an Os object.

    Args:
        info_text (str): The raw text output from Windows system information

    Returns:
        Os: An Os object containing the parsed Windows information
    """
    result = {}

    # Split the text into lines and process each line
    for line in info_text.strip().split("\n"):
        # Skip empty lines
        if not line.strip():
            continue

        # Split by colon and handle the key-value pairs
        parts = line.split(":", 1)
        if len(parts) == 2:
            key = parts[0].strip().replace("\x1b[32;1m", "").replace("\x1b[0m", "")
            value = parts[1].strip().replace("\x1b[32;1m", "").replace("\x1b[0m", "")
            result[key] = value

    # Extract relevant information for the Os object
    return Os(
        name=result.get("WindowsProductName", "Windows"),
        version=result.get("WindowsEditionId", ""),
        codename=result.get("WindowsProductId", ""),
        release=result.get("WindowsBuildLabEx", ""),
    )


def parse_macos_info(info_text) -> Os:
    """
    Parse macOS system information text into an Os object.

    Args:
        info_text (str): The raw text output from macOS system information

    Returns:
        Os: An Os object containing the parsed macOS information
    """
    result = {}

    # Split the text into lines and process each line
    for line in info_text.strip().split("\n"):
        # Skip empty lines
        if not line.strip():
            continue

        # Split by colon and handle the key-value pairs
        parts = line.split(":", 1)
        if len(parts) == 2:
            key = parts[0].strip()
            value = parts[1].strip()
            result[key] = value

    # Extract relevant information for the Os object
    return Os(
        name=result.get("ProductName", "macOS"),
        version=result.get("ProductVersion", ""),
        codename=result.get("BuildVersion", ""),
        release=result.get("ProductName", "macOS"),
    )


def parse_windows_updates(info_text) -> list[AppPackage]:
    """
    Parse Windows update information text into a list of dictionaries.

    Args:
        info_text (str): The raw text output from Windows update information

    Returns:
        list: A list of dictionaries containing the parsed Windows update information
    """
    updates = []

    # Split the text into lines
    lines = info_text.strip().split("\n")

    # Skip the header lines (first two lines)
    if len(lines) > 2:
        for line in lines[2:]:
            # Skip empty lines
            if not line.strip():
                continue

            # Split the line by whitespace, but keep the title together
            parts = line.split()

            # Ensure we have at least 5 parts (ComputerName, Status, KB, Size, Title)
            if len(parts) >= 5:
                # Extract the KB number
                kb = parts[2]

                # Extract the size
                size = parts[3]

                # The title is everything after the size
                title = " ".join(parts[4:])

                # Create a dictionary for this update
                update = AppPackage(
                    name=kb,
                    version=size,
                    release=title,
                    codename="",
                    description="",
                )

                updates.append(update)

    return updates


def parse_macos_updates(info_text) -> list[AppPackage]:
    """
    Parse macOS update information text into a list of UpdatePackage objects.

    Args:
        info_text (str): The raw text output from macOS update information

    Returns:
        list[UpdatePackage]: A list of UpdatePackage objects containing the parsed macOS update information
    """
    updates = []

    # Split the text into lines
    lines = info_text.strip().split("\n")

    # Find the line that indicates the start of the update list
    start_index = 0
    for i, line in enumerate(lines):
        if "Software Update found the following new or updated software:" in line:
            start_index = i + 1
            break

    # Process each update entry
    current_update = {}
    for line in lines[start_index:]:
        # Skip empty lines
        if not line.strip():
            continue

        # Check if this is a new update entry (starts with *)
        if line.strip().startswith("*"):
            # If we have a previous update, add it to the list
            if current_update:
                updates.append(
                    AppPackage(
                        name=current_update.get("label", ""),
                        version=current_update.get("version", ""),
                        release=current_update.get("title", ""),
                        codename=current_update.get("label", ""),
                        description=current_update.get("size", ""),
                    )
                )

            # Start a new update entry
            current_update = {}

            # Extract the label
            label_parts = line.strip().split(":", 1)
            if len(label_parts) == 2:
                current_update["label"] = label_parts[1].strip()
        else:
            # This is a continuation of the current update
            # Extract key-value pairs
            parts = line.strip().split(",")
            for part in parts:
                if ":" in part:
                    key, value = part.split(":", 1)
                    key = key.strip()
                    value = value.strip()

                    if key == "Title":
                        current_update["title"] = value
                    elif key == "Version":
                        current_update["version"] = value
                    elif key == "Size":
                        current_update["size"] = value

    # Add the last update if there is one
    if current_update:
        updates.append(
            AppPackage(
                name=current_update.get("label", ""),
                version=current_update.get("version", ""),
                release=current_update.get("title", ""),
                codename=current_update.get("label", ""),
                description=current_update.get("size", ""),
            )
        )

    return updates


def parse_debian_updates(info_text) -> list[AppPackage]:
    """
    Parse Debian Linux update information text into a list of UpdatePackage objects.

    Args:
        info_text (str): The raw text output from Debian update information

    Returns:
        list[UpdatePackage]: A list of UpdatePackage objects containing the parsed Debian update information
    """
    updates = []

    # Split the text into lines
    lines = info_text.strip().split("\n")

    # Process each line
    for line in lines:
        # Skip empty lines
        if not line.strip():
            continue

        # Split the line by "/" to separate package name and version info
        parts = line.split("/", 1)
        if len(parts) == 2:
            package_name = parts[0].strip()
            version_info = parts[1].strip()

            # Split the version info by spaces
            version_parts = version_info.split()
            if len(version_parts) >= 2:
                version = version_parts[0]
                codename = version_parts[1]

                # The architecture is the last part
                architecture = version_parts[-1] if len(version_parts) > 2 else ""

                # Create an UpdatePackage object for this update
                update = AppPackage(
                    name=package_name.replace("~", ""),
                    version=codename.replace("~", ""),
                    release=package_name.replace("~", ""),
                    codename=version.replace("~", ""),
                    description=architecture.replace("~", ""),
                )

                updates.append(update)

    return updates


def parse_windows_installed_apps(info_text) -> list[AppPackage]:
    """
    Parse Windows installed applications information text into a list of AppPackage objects.

    Args:
        info_text (str): The raw text output from Windows installed applications information

    Returns:
        list[AppPackage]: A list of AppPackage objects containing the parsed Windows installed applications information
    """
    apps = []

    # Split the text into lines
    lines = info_text.strip().split("\n")

    # Skip the header line (first line)
    if len(lines) > 1:
        for line in lines[1:]:
            # Skip empty lines
            if not line.strip():
                continue

            # Process each line, splitting by multiple spaces
            parts = []
            current_part = []
            spaces = 0

            for char in line:
                if char == " ":
                    spaces += 1
                    if (
                        spaces >= 2
                    ):  # If we hit multiple spaces, consider it a delimiter
                        if current_part:  # If we have collected some characters
                            parts.append("".join(current_part).strip())
                            current_part = []
                    else:
                        current_part.append(char)
                else:
                    spaces = 0
                    current_part.append(char)

            # Add the last part if there is one
            if current_part:
                parts.append("".join(current_part).strip())

            # Ensure we have at least 4 parts (Caption, InstallState, Name, Vendor, Version)
            if len(parts) >= 5:
                caption = parts[0]
                install_state = parts[1]
                name = parts[2]
                vendor = parts[3]
                version = parts[4]

                # Create an AppPackage object for this app
                app = AppPackage(
                    name=name,
                    version=version,
                    release=caption,
                    codename=install_state,
                    description=vendor,
                )

                apps.append(app)

    return apps


def parse_macos_installed_apps(info_text) -> list[AppPackage]:
    """
    Parse macOS installed applications information from JSON into a list of AppPackage objects.

    Args:
        info_text (str): The JSON output from macOS system_profiler SPApplicationsDataType -json

    Returns:
        list[AppPackage]: A list of AppPackage objects containing the parsed macOS installed applications information
    """
    apps = []

    try:
        import json

        # Parse the JSON string into a Python dictionary
        data = json.loads(info_text)

        # Extract the applications list
        applications = data.get("SPApplicationsDataType", [])

        for app in applications:
            name = app.get("_name", "")
            version = app.get("version", "")
            arch_kind = app.get("arch_kind", "")
            obtained_from = app.get("obtained_from", "")

            # Create an AppPackage object for this app
            app_package = AppPackage(
                name=name,
                version=version,
                release=arch_kind,
                codename=obtained_from,
                description="-",
            )

            apps.append(app_package)

    except json.JSONDecodeError as e:
        print(f"Error parsing JSON: {e}")
    except Exception as e:
        print(f"Error processing applications data: {e}")

    return apps


def parse_debian_installed_apps(info_text) -> list[AppPackage]:
    """
    Parse Debian Linux installed applications information text into a list of AppPackage objects.

    Args:
        info_text (str): The raw text output from Debian installed applications information

    Returns:
        list[AppPackage]: A list of AppPackage objects containing the parsed Debian installed applications information
    """
    apps = []

    # Split the text into lines
    lines = info_text.strip().split("\n")

    # Process each line
    for line in lines:
        # Skip empty lines
        if not line.strip():
            continue

        # Split the line by tab or multiple spaces
        parts = line.split("\t")
        if len(parts) == 1:
            # If no tabs, try splitting by multiple spaces
            parts = []
            current_part = []
            spaces = 0

            for char in line:
                if char == " ":
                    spaces += 1
                    if (
                        spaces >= 2
                    ):  # If we hit multiple spaces, consider it a delimiter
                        if current_part:  # If we have collected some characters
                            parts.append("".join(current_part).strip())
                            current_part = []
                    else:
                        current_part.append(char)
                else:
                    spaces = 0
                    current_part.append(char)

            # Add the last part if there is one
            if current_part:
                parts.append("".join(current_part).strip())

        # Ensure we have at least 2 parts (package name and version)
        if len(parts) >= 2:
            package_name = parts[0].strip()
            version = parts[1].strip()

            # Create an AppPackage object for this app
            app = AppPackage(
                name=package_name.replace("~", ""),
                version=version.replace("~", ""),
                release=package_name.replace("~", ""),
                codename="debian",
                description="",
            )

            apps.append(app)

    return apps



from abc import ABC, abstractmethod
from typing import Any, Optional, Dict


class ContentMessage(ABC):
    @abstractmethod
    def subject(self) -> str:
        """Get the subject of the message"""
        pass

    @abstractmethod
    def copy(self) -> "ContentMessage":
        """Create a copy of the message"""
        pass

    @abstractmethod
    def to_dict(self) -> Dict[str, Any]:
        """Convert message to dictionary format"""
        pass

    @abstractmethod
    def get(self, key: Optional[str] = None) -> Any:
        """Get value from message content"""
        pass



from pd_ai_agent_core.core_types.content_message import ContentMessage
from pd_ai_agent_core.messages.message import Message
import uuid
from typing import Optional
from pd_ai_agent_core.common.constants import ERROR_SUBJECT
from pd_ai_agent_core.common.message_status import MessageStatus


class ErrorMessage(ContentMessage):
    def __init__(
        self,
        error_message: str,
        error_type: Optional[str] = None,
        traceback: Optional[str] = None,
    ):
        self._subject = ERROR_SUBJECT
        self._content = {
            "error_message": error_message,
            "error_type": error_type,
            "traceback": traceback,
        }

    def copy(self) -> "ErrorMessage":
        return ErrorMessage(
            error_message=self._content["error_message"],
            error_type=self._content["error_type"],
            traceback=self._content["traceback"],
        )

    def subject(self) -> str:
        return self._subject

    def to_dict(self) -> dict:
        return {
            "error_message": self._content["error_message"],
            "error_type": self._content["error_type"],
            "traceback": self._content["traceback"],
        }

    def get(self, key: str) -> str:
        return self._content[key]


def create_error_message(
    session_id: str,
    channel: Optional[str],
    error_message: str,
    linked_message_id: Optional[str] = None,
    is_partial: bool = False,
) -> Message:
    msg = Message(
        session_id=session_id,
        channel=channel if channel is not None else str(uuid.uuid4()),
        subject=ERROR_SUBJECT,
        body=ErrorMessage(error_message),
    )
    msg.is_complete = True
    msg.status = MessageStatus.COMPLETE
    msg.linked_message_id = linked_message_id
    msg.is_partial = is_partial
    return msg


def create_error_message_from_message(
    message: Message,
    error_message: str,
    error_type: Optional[str] = None,
    traceback: Optional[str] = None,
    linked_message_id: Optional[str] = None,
    is_partial: bool = False,
) -> Message:
    """Create error message from existing message"""
    msg = Message(
        session_id=message.session_id,
        channel=message.channel,
        subject="error",
        context=message.context,
        body=ErrorMessage(error_message, error_type, traceback),
    )
    msg.sender = message.sender
    msg.status = message.status
    msg.message_id = message.message_id
    msg.tool_calls = message.tool_calls
    msg._is_complete = message._is_complete
    msg.linked_message_id = linked_message_id
    msg.is_partial = is_partial
    return msg



from pd_ai_agent_core.parallels_desktop.get_vm_screenshot import get_vm_screenshot
from pd_ai_agent_core.parallels_desktop.models.get_vm_result import GetVmResult
from pd_ai_agent_core.parallels_desktop.models.get_vms_result import GetVmsResult
from pd_ai_agent_core.parallels_desktop.datasource import VirtualMachineDataSource


def get_vms(take_screenshot: bool = False) -> GetVmResult:
    datasource = VirtualMachineDataSource.get_instance()
    vms = datasource.get_all_vms()
    if take_screenshot:
        for vm in vms:
            vm_screenshot_result = get_vm_screenshot(vm.id)
            vm.screenshot = vm_screenshot_result.screenshot or ""
            datasource.update_vm(vm)
    updated_vms = datasource.get_all_vms()
    updated_vms_dict = [vm.to_dict() for vm in updated_vms]
    return GetVmsResult(
        success=True,
        message="VMs listed successfully",
        exit_code=0,
        error="",
        raw_result=updated_vms_dict,
        vms=updated_vms,
    )


def get_vm(vm_id: str, take_screenshot: bool = False) -> GetVmResult:
    datasource = VirtualMachineDataSource.get_instance()
    vm = datasource.get_vm(vm_id)
    if vm is None:
        return GetVmResult(
            success=False,
            message="VM not found",
            exit_code=1,
            error="VM not found",
            vm=None,
        )
    if take_screenshot:
        vm_screenshot_result = get_vm_screenshot(vm.id)
        vm.screenshot = vm_screenshot_result.screenshot or ""
        datasource.update_vm(vm)
    updated_vm = datasource.get_vm(vm_id)
    if updated_vm is None:
        return GetVmResult(
            success=False,
            message="VM not found",
            exit_code=1,
            error="VM not found",
            vm=None,
        )
    updated_vm_dict = updated_vm.to_dict()
    return GetVmResult(
        success=True,
        message="VM listed successfully",
        exit_code=0,
        error="",
        raw_result=updated_vm_dict,
        vm=updated_vm,
    )



# Import and expose key modules
from .content_message import ContentMessage
from .background_agent import BackgroundAgent
from .llm_chat_ai_agent import LlmChatAgent, AgentFunction
from .session import Session
from .session_channel import SessionChannel
from .session_context import SessionContext
from .session_service import SessionService
from .session_type import SessionType

# Also expose the parallels_chat_ai_agent module for backward compatibility
from .llm_chat_ai_agent import LlmChatAgent as ParallelsLlmChatAgent
from .llm_chat_ai_agent import AgentFunction as ParallelsAgentFunction

__all__ = [
    "ContentMessage",
    "BackgroundAgent",
    "LlmChatAgent",
    "AgentFunction",
    "Session",
    "SessionChannel",
    "SessionContext",
    "SessionService",
    "SessionType",
    "ParallelsLlmChatAgent",
    "ParallelsAgentFunction",
]



from pd_ai_agent_core.messages.message import Message
from pd_ai_agent_core.core_types.content_message import ContentMessage
from typing import Dict, Any, Optional, List
import uuid
from pd_ai_agent_core.common.constants import (
    CHAT_SUBJECT,
    TOOL_CHANGE_SUBJECT,
    AGENT_FUNCTION_CALL_SUBJECT,
    GLOBAL_CHANNEL,
)
from pd_ai_agent_core.common.message_status import MessageStatus
from pydantic import BaseModel
from pd_ai_agent_core.core_types.llm_chat_ai_agent import AttachmentContextVariable


class StreamChatMessageAction(BaseModel):
    id: str
    name: str
    description: str
    type: str
    value: str
    parameters: dict
    icon: Optional[str] = None
    group_id: Optional[str] = None

    def to_dict(self):
        return {
            "id": self.id,
            "name": self.name,
            "description": self.description,
            "type": self.type,
            "value": self.value,
            "parameters": self.parameters,
            "icon": self.icon,
            "group_id": self.group_id,
        }


class StreamChatMessage(ContentMessage):
    def __init__(
        self,
        sender: str,
        role: str,
        content: str,
        actions: List[StreamChatMessageAction] = [],
        icon: Optional[str] = None,
        attachments: List[AttachmentContextVariable] = [],
    ):
        self.sender = sender
        self.role = role
        self.content = content
        self.msg_type = "chat_stream"
        self.icon = icon
        self.actions = actions
        self.attachments = attachments

    def to_dict(self) -> Dict[str, Any]:
        return {
            "sender": self.sender,
            "role": self.role,
            "content": self.content,
            "msg_type": self.msg_type,
            "actions": [action.to_dict() for action in self.actions],
            "icon": self.icon,
            "attachments": [attachment.to_dict() for attachment in self.attachments],
        }

    def copy(self) -> "StreamChatMessage":
        return StreamChatMessage(
            self.sender, self.role, self.content, self.actions, self.icon
        )

    def subject(self) -> str:
        return CHAT_SUBJECT

    def get(self, key: Optional[str] = None) -> Any:
        if key == "content":
            return self.content
        if key == "sender":
            return self.sender
        if key == "role":
            return self.role
        if key == "msg_type":
            return self.msg_type
        if key == "icon":
            return self.icon
        if key == "actions":
            return self.actions
        return None


class ToolChangeChatMessage(ContentMessage):
    def __init__(
        self, name: str, arguments: Dict[str, Any], icon: Optional[str] = None
    ):
        self.name = name
        self.arguments = arguments
        self.msg_type = "tool_change"
        self.icon = icon

    def to_dict(self) -> Dict[str, Any]:
        return {
            "name": self.name,
            "arguments": self.arguments,
            "msg_type": self.msg_type,
            "icon": self.icon,
        }

    def copy(self) -> "ToolChangeChatMessage":
        return ToolChangeChatMessage(self.name, self.arguments, self.icon)

    def subject(self) -> str:
        return TOOL_CHANGE_SUBJECT

    def get(self, key: Optional[str] = None) -> Any:
        if key == "name":
            return self.name
        if key == "arguments":
            return self.arguments
        if key == "msg_type":
            return self.msg_type
        if key == "icon":
            return self.icon
        if self.arguments is not None:
            for key in self.arguments:
                if key == key:
                    return self.arguments[key]
        return None


class AgentFunctionCallChatMessage(ContentMessage):
    def __init__(
        self, name: str, arguments: Dict[str, Any], icon: Optional[str] = None
    ):
        self.name = name
        self.arguments = arguments
        self.msg_type = "agent_function_call"
        self.icon = icon

    def to_dict(self) -> Dict[str, Any]:
        return {
            "name": self.name,
            "arguments": self.arguments,
            "msg_type": self.msg_type,
            "icon": self.icon,
        }

    def copy(self) -> "AgentFunctionCallChatMessage":
        return AgentFunctionCallChatMessage(self.name, self.arguments, self.icon)

    def subject(self) -> str:
        return AGENT_FUNCTION_CALL_SUBJECT

    def get(self, key: Optional[str] = None) -> Any:
        if key == "name":
            return self.name
        if key == "arguments":
            return self.arguments
        if key == "msg_type":
            return self.msg_type
        if key == "icon":
            return self.icon
        if self.arguments is not None:
            for key in self.arguments:
                if key == key:
                    return self.arguments[key]
        return None


def create_stream_chat_message(
    session_id: str,
    channel: Optional[str],
    sender: str,
    role: str,
    content: str,
    is_complete: bool,
    linked_message_id: Optional[str] = None,
    is_partial: bool = False,
    actions: List[StreamChatMessageAction] = [],
    icon: Optional[str] = None,
) -> Message:
    msg = Message(
        session_id=session_id,
        channel=channel if channel is not None else GLOBAL_CHANNEL,
        subject=CHAT_SUBJECT,
        icon=icon,
        body=StreamChatMessage(
            sender=sender,
            role=role,
            content=content,
            actions=actions,
            icon=icon,
        ),
    )
    msg.is_complete = is_complete
    msg.message_id = str(uuid.uuid4())
    if is_complete:
        msg.status = MessageStatus.COMPLETE
    else:
        msg.status = MessageStatus.STREAMING
    msg.linked_message_id = linked_message_id
    msg.is_partial = is_partial
    return msg


def create_tool_change_chat_message(
    session_id: str,
    channel: Optional[str],
    name: str,
    arguments: Dict[str, Any],
    linked_message_id: Optional[str] = None,
    is_partial: bool = False,
    icon: Optional[str] = None,
) -> Message:
    msg = Message(
        session_id=session_id,
        channel=channel if channel is not None else GLOBAL_CHANNEL,
        subject=TOOL_CHANGE_SUBJECT,
        icon=icon,
        body=ToolChangeChatMessage(
            name=name,
            arguments=arguments,
            icon=icon,
        ),
    )
    msg.is_complete = False
    msg.message_id = str(uuid.uuid4())
    msg.linked_message_id = linked_message_id
    msg.is_partial = is_partial
    return msg


def create_agent_function_call_chat_message(
    session_id: str,
    channel: Optional[str],
    name: str,
    arguments: Dict[str, Any] = {},
    linked_message_id: Optional[str] = None,
    is_partial: bool = False,
    icon: Optional[str] = None,
) -> Message:
    msg = Message(
        session_id=session_id,
        channel=channel if channel is not None else GLOBAL_CHANNEL,
        subject=AGENT_FUNCTION_CALL_SUBJECT,
        icon=icon,
        body=AgentFunctionCallChatMessage(
            name=name,
            arguments=arguments,
            icon=icon,
        ),
    )
    msg.is_complete = False
    msg.linked_message_id = linked_message_id
    msg.is_partial = is_partial
    msg.message_id = str(uuid.uuid4())
    return msg


def create_clean_agent_function_call_chat_message(
    session_id: str,
    channel: Optional[str],
    linked_message_id: Optional[str] = None,
    is_partial: bool = False,
) -> Message:
    msg = Message(
        session_id=session_id,
        channel=channel if channel is not None else GLOBAL_CHANNEL,
        subject=AGENT_FUNCTION_CALL_SUBJECT,
        body=AgentFunctionCallChatMessage(
            name="",
            arguments={},
        ),
    )
    msg.is_complete = False
    msg.linked_message_id = linked_message_id
    msg.is_partial = is_partial
    msg.message_id = str(uuid.uuid4())
    return msg



from abc import ABC, abstractmethod


class SessionService(ABC):
    def __init__(self, session_id: str):
        self.session_id = session_id

    @abstractmethod
    def name(self) -> str:
        pass

    @abstractmethod
    def register(self) -> None:
        pass

    @abstractmethod
    def unregister(self) -> None:
        pass



import cv2
import numpy as np
from collections import Counter
import base64
import logging
import io

logger = logging.getLogger(__name__)

# Try to import PIL - this will be needed for optimize_image
try:
    from PIL import Image

    PILLOW_AVAILABLE = True
except ImportError:
    PILLOW_AVAILABLE = False
    logger.warning("PIL not available, image optimization functions will be limited")


def optimize_image(
    image_data: str,
    quality: int = 60,
    max_width: int = 800,
    format: str = "JPEG",
    debug: bool = False,
) -> str:
    """
    Optimize an image by resizing and compressing it.

    Args:
        image_data (str): Base64 encoded image data
        quality (int): Output image quality (0-100, JPEG only)
        max_width (int): Maximum width in pixels (maintains aspect ratio)
        format (str): Output format ('JPEG', 'PNG', etc.)
        debug (bool): Whether to log debug information

    Returns:
        str: Base64 encoded optimized image
    """
    if not PILLOW_AVAILABLE:
        logger.warning("PIL not available, cannot optimize image")
        return image_data

    try:
        # Decode base64 to bytes
        image_bytes = base64.b64decode(image_data)

        # Convert to PIL Image
        img = Image.open(io.BytesIO(image_bytes))

        # Calculate new dimensions maintaining aspect ratio
        width, height = img.size
        if width > max_width:
            ratio = max_width / width
            new_width = max_width
            new_height = int(height * ratio)
            img = img.resize((new_width, new_height), Image.Resampling.LANCZOS)

        # Convert to desired format with compression
        output = io.BytesIO()
        if format.upper() == "JPEG":
            img = img.convert("RGB")  # JPEG doesn't support alpha channel
            img.save(output, format=format, quality=quality, optimize=True)
        else:
            img.save(output, format=format, optimize=True)

        output.seek(0)

        # Convert back to base64
        optimized_base64 = base64.b64encode(output.read()).decode("utf-8")

        # Log size reduction
        if debug:
            original_size = len(image_data)
            new_size = len(optimized_base64)
            reduction = (1 - (new_size / original_size)) * 100
            logger.info(
                f"Image optimized: {original_size:,} bytes → {new_size:,} bytes ({reduction:.1f}% reduction)"
            )

        return optimized_base64
    except Exception as e:
        logger.error(f"Error optimizing image: {e}")
        return image_data


def get_dominant_border_color(
    image_data: str, border_percentage: float = 0.2, debug: bool = False
) -> str:
    """
    Detect the most prevalent color in the borders of an image.

    Args:
        image (np.ndarray): Input image (BGR format).
        border_percentage (float): Percentage of the border to consider (0.0 to 1.0).

    Returns:
        tuple: Dominant BGR color (B, G, R).
    """
    imageBytes = base64.b64decode(image_data)
    np_array = np.frombuffer(imageBytes, dtype=np.uint8)
    image = cv2.imdecode(np_array, cv2.IMREAD_COLOR)

    if image is None:
        return "#F2F2F7"

    if len(image.shape) == 2:  # Grayscale image
        image = cv2.cvtColor(image, cv2.COLOR_GRAY2BGR)

    height, width, _ = image.shape
    border_thickness_h = int(height * border_percentage)
    border_thickness_w = int(width * border_percentage)

    top_border = image[:border_thickness_h, :]
    bottom_border = image[-border_thickness_h:, :]
    left_border = image[:, :border_thickness_w]
    right_border = image[:, -border_thickness_w:]

    borders_combined = np.concatenate(
        (
            top_border.reshape(-1, 3),
            bottom_border.reshape(-1, 3),
            left_border.reshape(-1, 3),
            right_border.reshape(-1, 3),
        ),
        axis=0,
    )

    # Count the most common color
    pixel_counts = Counter(map(tuple, borders_combined))
    dominant_color = pixel_counts.most_common(1)[0][0]

    hex_color = "#%02X%02X%02X" % dominant_color

    if debug:
        logger.info(f"Dominant Border Color: {hex_color}")

    return hex_color


def detect_black_screen(
    image_data: str,
    black_threshold: int = 30,
    black_percentage: float = 0.9,
    debug: bool = False,
) -> bool:
    """
    Detect if an image is mostly black.

    Args:
        image (np.ndarray): The input image (BGR or grayscale).
        black_threshold (int): Pixel values below this are considered black (0-255).
        black_percentage (float): Percentage threshold to classify as mostly black (0.0 - 1.0).

    Returns:
        bool: True if the image is mostly black, False otherwise.
    """
    imageBytes = base64.b64decode(image_data)
    image = np.frombuffer(imageBytes, dtype=np.uint8)

    if len(image.shape) == 3:
        gray = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
    else:
        gray = image

    black_pixels_mask = gray < black_threshold

    black_pixels_count = np.sum(black_pixels_mask)
    total_pixels = gray.size

    black_ratio = black_pixels_count / total_pixels

    if debug:
        logger.info(
            f"Black Pixels: {black_pixels_count}, Total Pixels: {total_pixels}, Black Ratio: {black_ratio:.2%}"
        )

    return bool(black_ratio >= black_percentage)


def scale_image(image_data: str, scale_factor: float = 0.5) -> str:
    """
    Scale an image by a given factor.
    """
    imageBytes = base64.b64decode(image_data)
    image = np.frombuffer(imageBytes, dtype=np.uint8)
    image = cv2.resize(image, None, fx=scale_factor, fy=scale_factor)

    # Convert back to base64
    optimized_base64 = base64.b64encode(image).decode("utf-8")

    return optimized_base64



import subprocess
import time
from typing import List
from pd_ai_agent_core.parallels_desktop.get_vms import get_vm
from pd_ai_agent_core.parallels_desktop.helpers import get_prlctl_command
from pd_ai_agent_core.parallels_desktop.models.execute_vm_command_result import (
    ExecuteVmCommandResult,
)


def execute_on_vm(
    vm_id: str, command: str, args: List[str] = []
) -> ExecuteVmCommandResult:
    if not vm_id:
        return ExecuteVmCommandResult(error="No VM ID provided", exit_code=1)
    try:
        vm_details = get_vm(vm_id)
        if not vm_details.success:
            return ExecuteVmCommandResult(
                error=vm_details.message, exit_code=vm_details.exit_code
            )
        if vm_details.vm.state != "running":
            return ExecuteVmCommandResult(error="VM is not running", exit_code=1)
    except Exception as e:
        return ExecuteVmCommandResult(error=str(e), exit_code=1)
    try:
        # testing if vm is available
        waitFor = 30
        is_available = False
        while waitFor > 0:
            helloArgs = []
            if vm_details.vm.os.lower().startswith("win"):
                helloArgs = ["print", "hello"]
            else:
                helloArgs = ["echo", "hello"]
            cmd: List[str] = [get_prlctl_command(), "exec", vm_id, *helloArgs]
            result = subprocess.run(
                cmd,
                capture_output=True,
                text=True,
                check=False,
                shell=False,
            )
            if result.returncode == 0:
                is_available = True
                break
            time.sleep(1)
            waitFor -= 1

        if not is_available:
            return ExecuteVmCommandResult(error="VM is not available", exit_code=1)

        cmdArgs = command.split(" ")
        if len(args) > 0:
            cmdArgs.extend(args)
        cmd: List[str] = [get_prlctl_command(), "exec", vm_id, *cmdArgs]
        result = subprocess.run(
            cmd,
            capture_output=True,
            text=True,
            check=False,
            shell=False,
        )
        if result.returncode != 0:
            return ExecuteVmCommandResult(
                error=result.stderr, exit_code=result.returncode
            )
        return ExecuteVmCommandResult(output=result.stdout, exit_code=result.returncode)
    except subprocess.CalledProcessError as e:
        return ExecuteVmCommandResult(error=str(e), exit_code=1)



import json
from typing import List, Callable, Union, Optional, Any, Dict
from pd_ai_agent_core.common.defaults import DEFAULT_MODEL
from pd_ai_agent_core.helpers.strings import normalize_string

# Third-party imports
from pydantic import BaseModel, HttpUrl
from enum import Enum

AgentFunction = Callable[[], Union[str, "LlmChatAgent", dict]]


class AttachmentType(Enum):
    """Enum representing different types of attachments that can be used in agent messages."""

    TEXT = "text"
    IMAGE = "image"
    CODEBLOCK = "codeblock"
    FILE = "file"
    AUDIO = "audio"
    VIDEO = "video"

    def __str__(self) -> str:
        return self.value


class AttachmentContextVariable(BaseModel):
    name: str
    id: str
    type: AttachmentType = AttachmentType.TEXT
    value: str
    download_url: Optional[str] = None
    language: Optional[str] = None
    format: Optional[str] = None

    def to_dict(self):
        return {
            "name": self.name,
            "id": self.id,
            "type": str(
                self.type
            ),  # Convert enum to string to make it JSON serializable
            "value": self.value,
            "download_url": self.download_url,
            "language": self.language,
            "format": self.format,
        }

    @classmethod
    def from_dict(cls, data: dict):
        return cls(
            name=data["name"],
            id=data["id"],
            type=AttachmentType(data["type"]),
            value=data["value"],
            download_url=data["download_url"],
            language=data["language"],
            format=data["format"],
        )


class DataResult:
    value: Any
    context_variables: Dict[str, Any] = dict()


class LlmChatAgentResponseAction(BaseModel):
    id: str
    name: str
    description: str
    type: str
    value: str
    parameters: dict
    icon: Optional[str] = None
    group_id: Optional[str] = None

    def to_dict(self):
        return {
            "id": self.id,
            "name": self.name,
            "description": self.description,
            "type": self.type,
            "value": self.value,
            "parameters": self.parameters,
            "icon": self.icon,
            "group_id": self.group_id,
        }

    @classmethod
    def from_dict(cls, data: dict):
        return cls(
            id=data["id"],
            name=data["name"],
            description=data["description"],
            type=data["type"],
            value=data["value"],
            parameters=data["parameters"],
            icon=data["icon"],
            group_id=data["group_id"],
        )


class AgentFunctionDescriptor:
    name: str
    description: str

    def __init__(self, name: str, description: str):
        self.name = name
        self.description = description

    def get_string(self) -> str:
        return f"{normalize_string(self.name)}::{self.description}"


class LlmChatAgent:
    def __init__(
        self,
        name: str,
        instructions: Union[str, Callable[[Dict[str, Any]], str]],
        description: Optional[str] = None,
        args: Optional[Dict[str, Any]] = None,
        env: Optional[Dict[str, Any]] = None,
        model: str = DEFAULT_MODEL,
        functions: List[AgentFunction] = [],
        function_descriptions: List[AgentFunctionDescriptor] = [],
        parallel_tool_calls: bool = True,
        transfer_instructions: Optional[str] = None,
        icon: Optional[Union[str, HttpUrl]] = None,
        tool_choice: Optional[str] = None,
    ):
        self.id = normalize_string(name)
        self.name = name
        self.description = description
        self.instructions = instructions
        self.model = model
        self.functions = functions or []
        self.function_descriptions = function_descriptions or []
        self.parallel_tool_calls = parallel_tool_calls
        self.transfer_instructions = transfer_instructions
        self.icon = icon
        self.tool_choice = tool_choice
        self.args = args
        self.env = env


class LlmChatResponse(BaseModel):
    messages: List = []
    agent: Optional[LlmChatAgent] = None
    context_variables: dict = {}
    actions: List[LlmChatAgentResponseAction] = []
    attachments: List[AttachmentContextVariable] = []

    class Config:
        arbitrary_types_allowed = True


class LlmChatResult(BaseModel):
    """
    Encapsulates the possible return values for an agent function.

    Attributes:
        value (str): The result value as a string.
        agent (Dict): The agent instance as a dictionary.
        context_variables (dict): A dictionary of context variables.
    """

    value: str = ""
    agent: Optional[LlmChatAgent] = None
    context_variables: dict = {}

    class Config:
        arbitrary_types_allowed = True


class LlmChatAgentResponse:
    status: str
    message: str
    error: Optional[str] = None
    data: Optional[Union[dict, List[dict]]] = None
    agent: Optional[LlmChatAgent] = None
    context_variables: dict = {}
    actions: List[LlmChatAgentResponseAction] = []
    attachments: List[AttachmentContextVariable] = []

    def __init__(
        self,
        status: str,
        message: str,
        error: Optional[str] = None,
        data: Optional[Union[dict, List[dict]]] = None,
        agent: Optional[LlmChatAgent] = None,
        context_variables: dict = {},
        actions: List[LlmChatAgentResponseAction] = [],
        attachments: List[AttachmentContextVariable] = [],
    ):
        self.status = status
        self.message = message
        self.error = error
        self.data = data
        self.agent = agent
        self.context_variables = context_variables
        self.actions = actions
        self.attachments = attachments

    def to_dict(self):
        return {
            "status": self.status,
            "message": self.message,
            "error": self.error,
            "data": self.data if self.data is not None else None,
            "actions": [action.to_dict() for action in self.actions],
            "attachments": [attachment.to_dict() for attachment in self.attachments],
        }

    def value(self) -> str:
        if self.data is not None:
            return json.dumps(self.data)
        if self.error:
            return self.error
        if self.message:
            return self.message
        return self.status

    @staticmethod
    def from_dict(data: dict):
        return LlmChatAgentResponse(
            status=data["status"],
            message=data["message"],
            error=data["error"],
            data=data["data"],
            actions=[
                LlmChatAgentResponseAction.from_dict(action)
                for action in data["actions"]
            ],
            attachments=[
                AttachmentContextVariable.from_dict(attachment)
                for attachment in data["attachments"]
            ],
        )



from .clone_vm_result import *
from .delete_vm_result import *
from .execute_vm_command_result import *
from .get_vm_os_version_result import *
from .get_vm_screenshot_result import *
from .operation_result import *
from .get_vm_result import *
from .set_vm_state_result import *
from .virtual_machine import *



from typing import Optional
from pd_ai_agent_core.common.constants import (
    NOTIFICATION_SERVICE_NAME,
    LOGGER_SERVICE_NAME,
)
from pd_ai_agent_core.messages.log_message import create_log_message, LogLevel
import uuid
import logging
from pd_ai_agent_core.helpers.strings import get_error_from_exception
from pd_ai_agent_core.services.service_registry import ServiceRegistry
from pd_ai_agent_core.services.notification_service import NotificationService
from pd_ai_agent_core.core_types.session_service import SessionService
import os
from pd_ai_agent_core.helpers.strings import parse_boolean

logger = logging.getLogger(__name__)


class LogService(SessionService):

    def __init__(self, session_id: str, debug: bool = False):
        self._debug = debug
        self._session_id = session_id
        ns = ServiceRegistry.get(
            session_id, NOTIFICATION_SERVICE_NAME, NotificationService
        )
        if ns is None:
            raise ValueError(f"Notification service not found for session {session_id}")
        self._ns = ns
        self._loop = None
        self._enable_ws_logging = False
        if os.getenv("PD_AI_DEBUG__SEND_WS_LOG") is not None:
            self._enable_ws_logging = parse_boolean(
                os.getenv("PD_AI_DEBUG__SEND_WS_LOG") or "false"
            )
        self.register()

    def name(self) -> str:
        return LOGGER_SERVICE_NAME

    def register(self) -> None:
        """Register this service with the registry"""
        if not ServiceRegistry.register(self._session_id, LOGGER_SERVICE_NAME, self):
            logger.info(
                f"Logger service already registered for session {self._session_id}"
            )
            return

    def unregister(self) -> None:
        """Unregister this service from the registry"""
        logger.info(f"Logger service unregistered for session {self._session_id}")

    def info(self, channel: Optional[str], message: str) -> None:
        log_message = create_log_message(
            session_id=self._session_id,
            channel=channel if channel else str(uuid.uuid4()),
            level=LogLevel.INFO,
            log_message=message,
        )
        logger.info(f"[{self._session_id}] [{channel}] {message}")
        if self._enable_ws_logging:
            self._ns.send_sync(log_message)

    def debug(self, channel: Optional[str], message: str) -> None:
        """Synchronous version of debug logging"""
        if not self._debug:
            return
        log_message = create_log_message(
            session_id=self._session_id,
            channel=channel if channel else str(uuid.uuid4()),
            level=LogLevel.DEBUG,
            log_message=message,
        )
        logger.debug(f"[{self._session_id}] [{channel}] {message}")
        if self._enable_ws_logging:
            self._ns.send_sync(log_message)

    def warning(self, channel: Optional[str], message: str) -> None:
        """Synchronous version of warning logging"""
        log_message = create_log_message(
            session_id=self._session_id,
            channel=channel if channel else str(uuid.uuid4()),
            level=LogLevel.WARNING,
            log_message=message,
        )
        logger.warning(f"[{self._session_id}] [{channel}] {message}")
        if self._enable_ws_logging:
            self._ns.send_sync(log_message)

    def error(self, channel: Optional[str], message: str) -> None:
        """Synchronous version of error logging"""

        log_message = create_log_message(
            session_id=self._session_id,
            channel=channel if channel else str(uuid.uuid4()),
            level=LogLevel.ERROR,
            log_message=message,
        )
        logger.error(f"[{self._session_id}] [{channel}] {message}")
        if self._enable_ws_logging:
            self._ns.send_sync(log_message)

    def exception(self, channel: Optional[str], message: str, e: Exception) -> None:
        """Synchronous version of error logging"""
        error_message = get_error_from_exception(message, e)
        log_message = create_log_message(
            session_id=self._session_id,
            channel=channel if channel else str(uuid.uuid4()),
            level=LogLevel.ERROR,
            log_message=f"{error_message['error_message']}: {error_message['error_type']}",
        )
        logger.error(f"[{self._session_id}] [{channel}] {message}")
        if self._enable_ws_logging:
            self._ns.send_sync(log_message)

    def trace(self, channel: Optional[str], message: str) -> None:
        """Synchronous version of trace logging"""
        log_message = create_log_message(
            session_id=self._session_id,
            channel=channel if channel else str(uuid.uuid4()),
            level=LogLevel.TRACE,
            log_message=message,
        )
        logger.debug(f"[{self._session_id}] [{channel}] {message}")
        if self._enable_ws_logging:
            self._ns.send_sync(log_message)



from typing import List
import subprocess
from pd_ai_agent_core.common.constants import NOTIFICATION_SERVICE_NAME
from pd_ai_agent_core.services.service_registry import ServiceRegistry
from pd_ai_agent_core.services.notification_service import NotificationService
from pd_ai_agent_core.messages.message import Message
from pd_ai_agent_core.messages.notification_message import (
    create_success_notification_message,
    create_error_notification_message,
)
from pd_ai_agent_core.parallels_desktop.helpers import get_prlctl_command
from pd_ai_agent_core.parallels_desktop.datasource import VirtualMachineDataSource
import logging
from pd_ai_agent_core.parallels_desktop.models.operation_result import OperationResult

logger = logging.getLogger(__name__)


async def reboot_vm(session_id: str, vm_id: str, message: Message) -> OperationResult:
    try:
        notifications = ServiceRegistry.get(
            session_id, NOTIFICATION_SERVICE_NAME, NotificationService
        )
        vm_datasource = VirtualMachineDataSource.get_instance()
        vm = vm_datasource.get_vm(vm_id)
        if vm:
            cmd: List[str] = [str(get_prlctl_command()), "restart", vm_id]
            result = subprocess.run(
                cmd,
                capture_output=True,
                text=True,
                check=True,
                shell=False,
            )
            if result.returncode == 0:
                success_message = create_success_notification_message(
                    session_id=session_id,
                    channel=vm_id,
                    message=f"VM {vm.name} restarted successfully",
                    details=f"The VM {vm.name} was restarted successfully",
                )
                opResult = OperationResult(
                    success=True,
                    message=f"VM {vm.name} restarted successfully",
                    exit_code=result.returncode,
                    error=result.stderr,
                )
                await notifications.send(success_message)
                return opResult
            else:
                error_message = create_error_notification_message(
                    session_id=session_id,
                    channel=vm_id,
                    message=f"Error restarting VM {vm.name}",
                    details=f"The VM {vm.name} was not restarted successfully",
                )
                opResult = OperationResult(
                    success=False,
                    message=f"Error restarting VM {vm.name}",
                    exit_code=result.returncode,
                    error=result.stderr,
                )
                await notifications.send(error_message)
                return opResult
        else:
            error_message = create_error_notification_message(
                session_id=session_id,
                channel=vm_id,
                message="VM not found",
                details="The VM was not found",
            )
            opResult = OperationResult(
                success=False,
                message="VM not found",
                exit_code=result.returncode,
                error=result.stderr,
            )
            await notifications.send(error_message)
            return opResult
    except Exception as e:
        logger.error(f"Error restarting VM: {e}")
        error_message = create_error_notification_message(
            session_id=session_id,
            channel=vm_id,
            message=f"Error restarting VM: {e}",
            details="The VM was not restarted successfully",
        )
        await notifications.send(error_message)
        opResult = OperationResult(
            success=False,
            message=f"Error restarting VM: {e}",
            exit_code=1,
            error=str(e),
        )
        return opResult



import pydantic
import json


class AppPackage(pydantic.BaseModel):
    name: str
    version: str
    release: str
    codename: str
    description: str

    def to_dict(self) -> dict:
        return {
            "name": self.name,
            "version": self.version,
            "release": self.release,
            "codename": self.codename,
            "description": self.description,
        }

    def to_string(self) -> str:
        return f"{self.name} {self.version} {self.release} {self.codename} {self.description}"

    def to_list(self) -> list[str]:
        return [self.name, self.version, self.release, self.codename, self.description]

    def to_json(self) -> str:
        return json.dumps(self.to_dict())


def updates_to_string(updates: list[AppPackage]) -> str:
    if len(updates) == 0:
        return "No updates found"
    return "\n".join([update.to_string() for update in updates])


def app_packages_to_csv(app_packages: list[AppPackage]) -> str:
    if len(app_packages) == 0:
        return "No installed apps found"
    headers = ["Name", "Version", "Release", "Codename", "Description"]
    rows = [headers] + [app_package.to_list() for app_package in app_packages]
    return "\n".join([",".join(row) for row in rows])


class UpdatePackageResponse(pydantic.BaseModel):
    updates: list[AppPackage]
    installed_apps: list[AppPackage]
    os: str
    vm_id: str
    warnings: list[str]
    errors: list[str]

    def to_dict(self) -> dict:
        return {
            "updates": [update.to_dict() for update in self.updates],
            "installed_apps": [app.to_dict() for app in self.installed_apps],
            "os": self.os,
            "vm_id": self.vm_id,
            "warnings": self.warnings,
            "errors": self.errors,
        }

    def has_errors(self) -> bool:
        return len(self.errors) > 0

    def has_warnings(self) -> bool:
        return len(self.warnings) > 0

    def has_updates(self) -> bool:
        return len(self.updates) > 0

    def has_installed_apps(self) -> bool:
        return len(self.installed_apps) > 0

    def to_string(self) -> str:
        result = f"OS: {self.os}\nVM ID: {self.vm_id}\nWarnings: {self.warnings}\nErrors: {self.errors}"
        if self.has_updates():
            result += f"\nUpdates: {updates_to_string(self.updates)}"
        if self.has_installed_apps():
            result += f"\nInstalled Apps: {updates_to_string(self.installed_apps)}"
        return result

    def installed_apps_to_json(self) -> str:
        if len(self.installed_apps) == 0:
            return "No installed apps found"
        return json.dumps([app.to_dict() for app in self.installed_apps])

    def updates_to_json(self) -> str:
        if len(self.updates) == 0:
            return "No updates found"
        return json.dumps([update.to_dict() for update in self.updates])

    def to_json(self) -> str:
        if len(self.updates) == 0 and len(self.installed_apps) == 0:
            return "No updates or installed apps found"
        return json.dumps(self.to_dict())



from pd_ai_agent_core.messages.message import Message
from pd_ai_agent_core.core_types.content_message import ContentMessage
from typing import Any, Optional, Dict
from pd_ai_agent_core.common.constants import CONFIG_SUBJECT
from pd_ai_agent_core.common.constants import GLOBAL_CHANNEL


class ConfigMessage(ContentMessage):
    def __init__(self, key: str, value: Any):
        self.content = {"key": key, "value": value}

    def copy(self) -> "ConfigMessage":
        return ConfigMessage(
            key=self.content["key"],
            value=self.content["value"],
        )

    def subject(self) -> str:
        return CONFIG_SUBJECT

    def get(self, key: Optional[str] = None) -> Any:
        if key is None:
            return self.content
        return self.content.get(key)

    def to_dict(self) -> Dict[str, Any]:
        return {"key": self.content["key"], "value": self.content["value"]}


def create_config_message(
    session_id: str,
    channel: Optional[str],
    key: str,
    value: Any,
    linked_message_id: Optional[str] = None,
    is_partial: bool = False,
) -> Message:
    msg = Message(
        session_id=session_id,
        channel=channel if channel is not None else GLOBAL_CHANNEL,
        subject=CONFIG_SUBJECT,
        body=ConfigMessage(key, value),
    )
    msg.linked_message_id = linked_message_id
    msg.is_partial = is_partial
    return msg



import os
import yaml
import json
from typing import Any, List
from pathlib import Path
from pd_ai_agent_core.core_types.session_service import SessionService
from pd_ai_agent_core.common.constants import CONFIG_SERVICE_NAME
from pd_ai_agent_core.services.service_registry import ServiceRegistry
from pd_ai_agent_core.common.defaults import USER_DATA_DIR
import logging

logger = logging.getLogger(__name__)


class ConfigService(SessionService):
    CONFIG_DIR = Path.home() / USER_DATA_DIR / "config"

    def __init__(
        self, session_id: str, config_file: str = "config.yaml", debug: bool = False
    ):
        self.config_file = self.CONFIG_DIR / f"{session_id}.{config_file}"
        self.config: dict = {}
        self._session_id = session_id
        self.debug = debug
        self._ensure_config_dir()
        self.load()
        self.register()

    def name(self) -> str:
        return CONFIG_SERVICE_NAME

    def register(self) -> None:
        """Register this service with the registry"""
        if not ServiceRegistry.register(self._session_id, CONFIG_SERVICE_NAME, self):
            logger.info(
                f"Config service already registered for session {self._session_id}"
            )
            return

    def unregister(self) -> None:
        """Unregister this service from the registry"""
        logger.info(f"Config service unregistered for session {self._session_id}")

    def _ensure_config_dir(self) -> None:
        """Ensure .parallels directory exists"""
        self.CONFIG_DIR.mkdir(parents=True, exist_ok=True)

    def _parse_nested_key(self, key: str) -> List[str]:
        """Convert nested key to parts (e.g., 'database.host' -> ['database', 'host'])"""
        return key.split(".")

    def _set_nested_value(
        self, data: dict | list, parts: List[str], value: Any
    ) -> None:
        """Set value in nested dictionary"""
        for i, part in enumerate(parts[:-1]):
            if part.isdigit():  # Handle list indices
                idx = int(part)
                if not isinstance(data, list):
                    data = []  # type: ignore
                while len(data) <= idx:
                    data.append({})
            else:
                if not isinstance(data, dict) or part not in data:
                    data[part] = {}  # type: ignore
            data = data[part]  # type: ignore
        data[parts[-1]] = value  # type: ignore

    def _get_nested_value(
        self, data: dict, parts: List[str], default: Any = None
    ) -> Any:
        """Get value from nested dictionary"""
        for part in parts:
            if part.isdigit() and isinstance(data, list):
                idx = int(part)
                if idx >= len(data):
                    return default
                data = data[idx]
            elif isinstance(data, dict):
                if part not in data:
                    return default
                data = data[part]
            else:
                return default
        return data

    def load(self) -> None:
        """Load configuration from file and environment"""
        if self.config_file.exists():
            with open(self.config_file) as f:
                if str(self.config_file).endswith(".yaml"):
                    self.config = yaml.safe_load(f) or {}
                else:
                    self.config = json.load(f)
        else:
            self.config = {}
            self.save()

        # Override with environment variables
        for key, value in os.environ.items():
            if key.startswith("APP_"):
                # Convert APP_DATABASE__HOST to database.host
                config_key = key[4:].lower().replace("__", ".")
                # Parse value type
                if value.lower() in (
                    "true",
                    "false",
                    "t",
                    "f",
                    "1",
                    "0",
                    "yes",
                    "no",
                    "y",
                    "n",
                    "on",
                    "off",
                    "enabled",
                    "disabled",
                    "enable",
                    "disable",
                ):
                    value = value.lower() == "true"
                elif value.replace(".", "").isdigit():
                    value = float(value) if "." in value else int(value)
                elif value.startswith("[") and value.endswith("]"):
                    value = json.loads(value)

                parts = self._parse_nested_key(config_key)
                self._set_nested_value(self.config, parts, value)

    def save(self) -> None:
        """Save current configuration to file"""
        with open(self.config_file, "w") as f:
            if str(self.config_file).endswith(".yaml"):
                yaml.safe_dump(self.config, f)
            else:
                json.dump(self.config, f, indent=2)

    def get_key(self, key: str, default: Any = None) -> Any:
        """Get value for key with optional default"""
        env_key = f"APP_{key.upper().replace('.', '__')}"
        if env_key in os.environ:
            value = os.environ[env_key]
            # Parse value type
            if value.lower() in ("true", "false"):
                return value.lower() == "true"
            elif value.replace(".", "").isdigit():
                return float(value) if "." in value else int(value)
            return value

        parts = self._parse_nested_key(key)
        return self._get_nested_value(self.config, parts, default)

    def get_str_key(self, key: str, default: str = "") -> str:
        """Get string value for key"""
        return str(self.get_key(key, default))

    def get_int_key(self, key: str, default: int = 0) -> int:
        """Get integer value for key"""
        value = self.get_key(key, default)
        try:
            return int(float(value))  # Handle float strings
        except (TypeError, ValueError):
            return default

    def get_bool_key(self, key: str, default: bool = False) -> bool:
        """Get boolean value for key"""
        value = self.get_key(key, default)
        if isinstance(value, bool):
            return value
        if isinstance(value, str):
            return value.lower() in (
                "true",
                "1",
                "yes",
                "on",
                "enable",
                "enabled",
                "t",
                "y",
            )
        return bool(value)

    def get_float_key(self, key: str, default: float = 0.0) -> float:
        """Get float value for key"""
        value = self.get_key(key, default)
        try:
            return float(value)
        except (TypeError, ValueError):
            return default

    def get_list_key(self, key: str, default: List[Any] | None = None) -> List[Any]:
        """Get list value for key"""
        default_list = default if default is not None else []
        value = self.get_key(key, default_list)
        if isinstance(value, str):
            try:
                return json.loads(value)
            except json.JSONDecodeError:
                return default_list
        return value if isinstance(value, list) else default_list

    def set_key(self, key: str, value: Any) -> None:
        """Set value for key and save configuration"""
        parts = self._parse_nested_key(key)
        self._set_nested_value(self.config, parts, value)
        self.save()



from typing import Optional, Any
from pd_ai_agent_core.core_types.content_message import ContentMessage
import uuid
from pd_ai_agent_core.common.constants import ERROR_SUBJECT
from pd_ai_agent_core.common.message_status import MessageStatus


class Message:
    def __init__(
        self,
        session_id: str,
        channel: str = "",
        subject: str = "",
        body: ContentMessage | dict = {},
        context: Optional[dict] = None,
        icon: Optional[str] = None,
    ):
        self.session_id = session_id
        self.channel = channel
        self.subject = subject
        self.body = body
        self.context = context
        self._is_complete = False
        self.message_id = str(uuid.uuid4())
        self.sender = None
        self.error = None
        self.status = None
        self._linked_message_id = None
        self._is_partial = False
        self._partial_index = -1
        self.icon = None
        self.tool_calls = None

    @property
    def is_chat(self) -> bool:
        """Check if message is a chat type"""
        return self.subject == "chat"

    @property
    def is_command(self) -> bool:
        """Check if message is a command type"""
        return self.subject == "command"

    @property
    def is_complete(self) -> bool:
        """Check if message processing is complete"""
        return self._is_complete

    @is_complete.setter
    def is_complete(self, value: bool) -> None:
        """Set message completion status"""
        self._is_complete = value

    @property
    def linked_message_id(self) -> Optional[str]:
        """Get linked message ID"""
        return self._linked_message_id

    @linked_message_id.setter
    def linked_message_id(self, value: Optional[str]) -> None:
        """Set linked message ID"""
        self._linked_message_id = value

    @property
    def is_partial(self) -> bool:
        """Check if message is a partial message"""
        return self._is_partial

    @is_partial.setter
    def is_partial(self, value: bool) -> None:
        """Set partial message status"""
        self._is_partial = value

    @property
    def partial_index(self) -> int:
        """Get partial index"""
        return self._partial_index

    @partial_index.setter
    def partial_index(self, value: int) -> None:
        """Set partial index"""
        self._partial_index = value

    def increment_partial_index(self) -> None:
        """Increment partial index"""
        self._partial_index += 1

    def create_with_error(self, error_message: str) -> "Message":
        """Create a new message instance with error from existing message"""
        body_copy = (
            self.body.copy() if isinstance(self.body, dict) else self.body.copy()
        )
        new_message = Message(
            session_id=self.session_id,
            channel=self.channel,
            subject=ERROR_SUBJECT,
            body=body_copy,
            context=self.context.copy() if self.context else None,
        )
        new_message.is_complete = True
        new_message.message_id = self.message_id
        new_message.sender = self.sender
        new_message.error = error_message
        return new_message

    def mark_complete(self) -> None:
        """Mark message as complete"""
        self._is_complete = True

    def set_message_id(self, message_id: str) -> None:
        """Set message ID for tracking stream responses"""
        self.message_id = message_id

    def set_sender(self, sender: str) -> None:
        """Set message sender"""
        self.sender = sender

    def get_body(self, key: Optional[str] = None):
        """Get message body, optionally accessing nested key"""
        if isinstance(self.body, dict):
            if key is None:
                return self.body
            return self.body.get(key)
        # Handle IbodyMessage
        return self.body.get(key) if key is None else self.body.get(key)

    def get_context(self, key: Optional[str] = None):
        """Get message context, optionally accessing nested key"""
        if self.context is None:
            return None
        if key is None:
            return self.context
        return self.context.get(key)

    def set_status(self, status: MessageStatus) -> None:
        """Set message status"""
        self.status = status

    def get_status(self) -> Optional[MessageStatus]:
        """Get message status"""
        return self.status

    def set_tool_calls(self, tool_calls: list) -> None:
        """Set tool calls for message"""
        self.tool_calls = tool_calls

    def get_tool_calls(self) -> Optional[list]:
        """Get tool calls for message"""
        return self.tool_calls

    def to_dict(self) -> dict:
        """Convert message to dictionary format"""
        body_dict = (
            self.body.to_dict() if isinstance(self.body, ContentMessage) else self.body
        )
        return {
            "session_id": self.session_id,
            "channel": self.channel,
            "message_id": self.message_id,
            "subject": self.subject,
            "body": body_dict,
            "is_complete": self.is_complete,
            "sender": self.sender,
            "error": self.error,
            "context": self.context,
            "status": self.status.value if self.status else None,
            "tool_calls": self.tool_calls,
            "linked_message_id": self.linked_message_id,
            "is_partial": self.is_partial,
            "partial_index": self.partial_index,
            "icon": self.icon,
        }

    @classmethod
    def from_dict(cls, data: dict) -> "Message":
        """Create message instance from dictionary"""
        status = data.get("status")
        msg = cls(
            session_id=data["session_id"],
            channel=data["channel"],
            subject=data["subject"],
            body=data["body"],
        )
        msg.context = data.get("context")

        msg.error = data.get("error")
        msg.is_complete = data.get("is_complete", False)
        msg.message_id = data.get("message_id")
        msg.sender = data.get("sender")
        msg.status = MessageStatus(status) if status else None
        msg.tool_calls = data.get("tool_calls")
        msg.linked_message_id = data.get("linked_message_id")
        msg.is_partial = data.get("is_partial", False)
        msg.partial_index = data.get("partial_index", -1)
        msg.icon = data.get("icon")
        return msg

    def copy(self) -> "Message":
        """Create a copy of the message"""
        body_copy = (
            self.body.copy() if isinstance(self.body, dict) else self.body.copy()
        )
        new_message = Message(
            session_id=self.session_id,
            channel=self.channel,
            subject=self.subject,
            body=body_copy,
            context=self.context.copy() if self.context else None,
        )
        new_message.is_complete = self.is_complete
        new_message.message_id = self.message_id
        new_message.sender = self.sender
        new_message.error = self.error
        new_message.status = self.status
        new_message.tool_calls = self.tool_calls
        new_message.linked_message_id = self.linked_message_id
        new_message.is_partial = self.is_partial
        new_message.partial_index = self.partial_index
        new_message.icon = self.icon
        return new_message

    def get(self, key: Optional[str] = None) -> Any:
        """Get value from message body"""
        if isinstance(self.body, dict):
            if key is None:
                return self.body
            return self.body.get(key)
        return self.body.get(key) if key is None else self.body.get(key)



from pd_ai_agent_core.core_types.content_message import ContentMessage
from typing import Any, Dict, Optional
from pd_ai_agent_core.messages.message import Message
from pd_ai_agent_core.common.constants import COMMAND_SUBJECT, GLOBAL_CHANNEL


class CommandMessage(ContentMessage):
    def __init__(
        self,
        command: str,
        args: Dict[str, Any],
        result: Optional[Any] = None,
        error: Optional[Any] = None,
        exit_code: Optional[int] = None,
        stdout: Optional[str] = None,
        stderr: Optional[str] = None,
    ):
        self._subject = COMMAND_SUBJECT
        self._content = {
            "command": command,
            "args": args,
            "result": result,
            "error": error,
            "exit_code": exit_code,
            "stdout": stdout,
            "stderr": stderr,
        }

    def copy(self) -> "CommandMessage":
        return CommandMessage(
            command=self._content["command"],
            args=self._content["args"],
            result=self._content["result"],
            error=self._content["error"],
            exit_code=self._content["exit_code"],
            stdout=self._content["stdout"],
            stderr=self._content["stderr"],
        )

    def subject(self) -> str:
        return self._subject

    def to_dict(self) -> Dict[str, Any]:
        return {
            "command": self._content["command"],
            "args": self._content["args"],
            "result": self._content["result"],
            "error": self._content["error"],
            "exit_code": self._content["exit_code"],
            "stdout": self._content["stdout"],
            "stderr": self._content["stderr"],
        }

    def get(self, key: Optional[str] = None) -> Any:
        if key is None:
            return self._content
        if key in self._content:
            return self._content[key]
        if "args" in self._content:
            return (
                self._content["args"][key] if key is not None else self._content["args"]
            )
        return None

    def command(self) -> str:
        return self._content["command"]

    def args(self) -> Dict[str, Any]:
        return self._content["args"]

    def result(self) -> Optional[Any]:
        return self._content["result"]

    def error(self) -> Optional[Any]:
        return self._content["error"]

    def exit_code(self) -> Optional[int]:
        return self._content["exit_code"]

    def stdout(self) -> Optional[str]:
        return self._content["stdout"]

    def stderr(self) -> Optional[str]:
        return self._content["stderr"]


def create_command_message(
    session_id: str,
    channel: Optional[str],
    command: str,
    result: Optional[Any] = None,
    error: Optional[Any] = None,
    exit_code: Optional[int] = None,
    stdout: Optional[str] = None,
    stderr: Optional[str] = None,
    args: Dict[str, Any] = {},
    linked_message_id: Optional[str] = None,
    is_partial: bool = False,
) -> Message:
    msg = Message(
        session_id=session_id,
        channel=channel if channel is not None else GLOBAL_CHANNEL,
        subject=COMMAND_SUBJECT,
        body=CommandMessage(
            command,
            args,
            result,
            error,
            exit_code,
            stdout,
            stderr,
        ),
    )
    msg.linked_message_id = linked_message_id
    msg.is_partial = is_partial
    return msg



from .background_service import *
from .config_service import *
from .log_service import *
from .notification_service import *
from .ocr_service import *
from .service_registry import *
from .session_manager import *
from .vm_event_monitor_service import *



import asyncio
import logging
from typing import Dict, List
from datetime import datetime, timedelta
from fnmatch import fnmatch
from pd_ai_agent_core.core_types.background_agent import BackgroundAgent
from pd_ai_agent_core.core_types.session_service import SessionService
from pd_ai_agent_core.messages.background_message import BackgroundMessage
from pd_ai_agent_core.common.constants import BACKGROUND_SERVICE_NAME
from collections import deque
from pd_ai_agent_core.services.service_registry import ServiceRegistry
import concurrent.futures

logger = logging.getLogger(__name__)


class BackgroundAgentService(SessionService):
    def __init__(
        self,
        session_id: str,
        rate_limit_messages: int = 100,
        rate_limit_window: int = 60,
        debug: bool = False,
    ):
        self._agents: Dict[tuple[str, str], BackgroundAgent] = (
            {}
        )  # (session_id, agent_type) -> agent
        self._tasks: Dict[tuple[str, str], asyncio.Task] = {}
        self._message_queue: deque[BackgroundMessage] = deque()
        self._processing = False
        self._session_id = session_id
        self._rate_limit_messages = rate_limit_messages
        self._rate_limit_window = rate_limit_window
        self._message_timestamps: deque[datetime] = deque()
        self.debug = debug
        self._thread_pool = concurrent.futures.ThreadPoolExecutor(max_workers=10)
        self._main_loop = asyncio.get_event_loop()
        self._agent_loops: Dict[tuple[str, str], asyncio.AbstractEventLoop] = {}
        self.register()

    def name(self) -> str:
        return BACKGROUND_SERVICE_NAME

    def register(self) -> None:
        """Register this service with the registry"""
        if not ServiceRegistry.register(
            self._session_id, BACKGROUND_SERVICE_NAME, self
        ):
            logger.info(
                f"Background service already registered for session {self._session_id}"
            )
            return

    def unregister(self) -> None:
        """Unregister this service from the registry"""
        self.unregister_background_agent()
        logger.info(f"Background service unregistered for session {self._session_id}")

    async def add_background_agent(self, agent: BackgroundAgent) -> bool:
        """Add a new background agent if one doesn't exist for this session"""
        key = (agent.session_id, agent.agent_type)
        if key in self._agents:
            return False

        self._agents[key] = agent
        # Create a new event loop for this agent
        if agent.interval:
            self._tasks[key] = asyncio.create_task(self._run_agent_timer(agent))

        # Ensure message processing is running
        if not self._processing:
            self._processing = True
            asyncio.create_task(self._process_message_queue())

        return True

    def remove_background_agent(self, session_id: str, agent_type: str) -> bool:
        """Remove an agent and cancel its task"""
        key = (session_id, agent_type)
        if key not in self._agents:
            # Try to find the agent by session_id and agent_className
            found = False
            for k, v in self._agents.items():
                if k[0] == session_id and v.agent_className == agent_type:
                    key = k
                    found = True
                    break
            if not found:
                return False

        if key in self._tasks:
            self._tasks[key].cancel()
            del self._tasks[key]

        del self._agents[key]
        return True

    def unregister_background_agent(self):
        """Unregister all agents and stop the background service"""
        self._processing = False

        # Create a list of agents to avoid modifying dict during iteration
        agents_to_remove = [
            (agent.session_id, agent.agent_type) for agent in self._agents.values()
        ]

        # Remove each agent
        for session_id, agent_type in agents_to_remove:
            self.remove_background_agent(session_id, agent_type)

        # Clear all remaining data structures
        self._agents.clear()
        self._tasks.clear()
        self._message_queue.clear()
        self._message_timestamps.clear()

    def post_message(self, message_type: str, data: dict) -> None:
        """Synchronous method to post a message to the queue"""
        if not self._check_rate_limit():
            logger.warning("Rate limit exceeded, message dropped")
            return

        message = BackgroundMessage(message_type=message_type, data=data)
        self._message_queue.append(message)
        self._message_timestamps.append(datetime.now())
        logger.info(f"Message queued: {message_type}")

    def _check_rate_limit(self) -> bool:
        """Check if we're within rate limits"""
        now = datetime.now()
        cutoff = now - timedelta(seconds=self._rate_limit_window)

        # Remove old timestamps
        while self._message_timestamps and self._message_timestamps[0] < cutoff:
            self._message_timestamps.popleft()

        return len(self._message_timestamps) < self._rate_limit_messages

    async def _run_agent_timer(self, agent: BackgroundAgent):
        """Run the agent's process function at specified intervals"""
        # key = (agent.session_id, agent.agent_type)
        while True:
            if agent.interval is None:
                await asyncio.sleep(0)
            else:
                try:
                    # Run agent process in thread pool
                    await self._main_loop.run_in_executor(
                        self._thread_pool, lambda: asyncio.run(agent.process())
                    )
                except Exception as e:
                    logger.error(
                        f"Error in agent {agent.agent_type} for session {agent.session_id}: {e}"
                    )
            await asyncio.sleep(agent.interval or 0)

    async def _process_message_queue(self):
        """Process messages from the queue"""
        while self._processing:
            if self._message_queue:
                message = self._message_queue.popleft()
                logger.info(f"Processing message: {message.message_type}")
                await self._distribute_message(message)
            await asyncio.sleep(0.1)

    async def _distribute_message(self, message: BackgroundMessage):
        """Distribute a message to all subscribed agents"""
        tasks = []

        for agent in self._agents.values():
            for pattern in agent.subscribed_messages:
                if fnmatch(message.message_type, pattern):
                    # Create task for each agent that should process the message
                    task = asyncio.create_task(
                        self._process_agent_message(agent, message)
                    )
                    tasks.append(task)
                    break  # Agent matched, no need to check other patterns

        if tasks:
            # Wait for all agent tasks to complete
            await asyncio.gather(*tasks)

    async def _process_agent_message(
        self, agent: BackgroundAgent, message: BackgroundMessage
    ):
        """Process a message for a single agent with retry logic in a separate thread"""

        async def process_with_retries():
            while message.retry_count < message.max_retries:
                try:
                    await agent.process_message(message)
                    return
                except Exception as e:
                    message.retry_count += 1
                    logger.error(
                        f"Error processing message {message.message_type} in agent {agent.agent_type} "
                        f"(attempt {message.retry_count}/{message.max_retries}): {e}"
                    )
                    if message.retry_count < message.max_retries:
                        await asyncio.sleep(1 * message.retry_count)

        try:
            # Run the processing in a thread pool
            await self._main_loop.run_in_executor(
                self._thread_pool, lambda: asyncio.run(process_with_retries())
            )
        except Exception as e:
            logger.error(f"Error in thread pool execution: {e}")

    def get_agents(self) -> List[BackgroundAgent]:
        """Get all registered agents"""
        return list(self._agents.values())



from dataclasses import dataclass
from typing import Any, Dict
from datetime import datetime
from pd_ai_agent_core.core_types.content_message import ContentMessage
from pd_ai_agent_core.common.constants import BACKGROUND_MESSAGE_SUBJECT


@dataclass
class BackgroundMessage(ContentMessage):
    message_type: str  # Supports wildcards like "vm.*"
    data: Dict[str, Any]
    timestamp: datetime = datetime.now()
    retry_count: int = 0
    max_retries: int = 10

    def __init__(self, message_type: str, data: Dict[str, Any]):
        self.message_type = message_type
        self.data = data
        self.timestamp = datetime.now()
        self.retry_count = 0
        self.max_retries = 10

    def subject(self) -> str:
        return BACKGROUND_MESSAGE_SUBJECT

    def to_dict(self) -> Dict[str, Any]:
        return {
            "message_type": self.message_type,
            "data": self.data,
        }

    def get(self, key: str) -> Any:
        return self.data.get(key)

    def copy(self) -> "BackgroundMessage":
        return BackgroundMessage(self.message_type, self.data)

    @staticmethod
    def from_dict(data: ContentMessage | dict[str, Any]) -> "BackgroundMessage | None":
        """Create a BackgroundMessage from a dictionary, returns None if invalid"""
        if isinstance(data, ContentMessage):
            data = data.to_dict()

        try:
            if not isinstance(data, dict):
                return None

            message_type = data.get("message_type")
            message_data = data.get("data", {})

            if not message_type or not isinstance(message_type, str):
                return None

            result = BackgroundMessage(
                message_type=message_type,
                data=message_data,
            )
            return result
        except Exception:
            return None


def create_background_message(
    message_type: str, data: Dict[str, Any]
) -> BackgroundMessage:
    return BackgroundMessage(message_type, data)



from pydantic import BaseModel
from .virtual_machine import VirtualMachine
from typing import List


class GetVmsResult(BaseModel):
    success: bool
    message: str
    exit_code: int
    error: str
    raw_result: List[dict]
    vms: List[VirtualMachine]

    class Config:
        arbitrary_types_allowed = True

    def to_dict(self) -> dict:
        return self.model_dump()



from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

# Define requirements directly
requirements = [
    "websockets>=13.0.0",
    "pydantic>=1.10.0,<2.0.0",
    "typing-extensions>=4.0.0",
    "semver>=3.0.0",
    "easyocr",
    "opencv-python",
    "matplotlib",
    "numpy",
]

setup(
    name="pd-ai-agent-core",
    version="0.1.14",
    author="Parallels",
    author_email="carlos.lapao@parallels.com",
    description="A core library for AI agent functionality",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Parallels/pd-ai-agent-core",
    packages=find_packages(
        include=["pd_ai_agent_core", "pd_ai_agent_core.*"], exclude=["tests*"]
    ),
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
    python_requires=">=3.8",
    install_requires=requirements,
)



VM_SECURITY_CHECKS = "vm.security_checks"
VM_SYNC_SCREENSHOT = "vm.sync_screenshot"
VM_STATE_CHANGED = "vm.state_changed"
VM_RUN_UPDATE = "vm.run_update"
VM_HEALTH_CHECK = "vm.health_check"
VM_REBOOT = "vm.reboot"
VM_DISABLE_HEALTH_CHECK_TEST = "vm.disable_health_check_test"
VM_SEND_REPORT = "vm.send_report"
VM_STATE_STARTED = "vm.state.started"
VM_STATE_STOPPED = "vm.state.stopped"
VM_STATE_SUSPENDED = "vm.state.suspended"
VM_STATE_RESUMED = "vm.state.resumed"
VM_STATE_PAUSED = "vm.state.paused"
VM_STATE_CHANGED = "vm.state.changed"
PACKAGE_UPDATE = "package.update"
SYSTEM_UPDATE = "system.update"
SYSTEM_ERROR = "system.error"
SYSTEM_WARNING = "system.warning"
SYSTEM_INFO = "system.info"
SYSTEM_DEBUG = "system.debug"
SYSTEM_TRACE = "system.trace"
SYSTEM_CRITICAL = "system.critical"
SYSTEM_RESTART_BACKEND = "system.restart_backend"
CHAT_CLEAR_HISTORY = "chat.clear_history"
CHAT_CLEAR_CONTEXT = "chat.clear_context"
CHAT_CLEAR_SESSION = "chat.clear_session"
CHAT_CLEAR_ALL = "chat.clear_all"
CHAT_CLEAR_ALL_MESSAGES = "chat.clear_all_messages"
CHAT_CLEAR_ALL_MESSAGES = "chat.clear_all_messages"
CONFIG_SETTING_CHANGED = "config.setting_changed"
